import pandas as pd
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split

import time
from datetime import datetime


class DatasetPreprocessing():
    @staticmethod
    def add_time_information(data):
        if 'timestamp' not in data:
            data['timestamp'] = datetime.now().timestamp()

        date = datetime.fromtimestamp(data['timestamp'])
        data['datetime'] = date.strftime('%Y-%m-%d %H:%M:%s')
        data['month_of_year'] = date.month
        data['day_of_month'] = date.day
        data['day_of_week'] = date.strftime('%A')
        data['hour_of_day'] = date.hour
        data['minute_of_hour'] = date.minute

        return data

    @staticmethod
    def prepare_training(modelcfg, df):
        if modelcfg.has_categoricals():
            x = pd.get_dummies(df[modelcfg.features],
                               columns=modelcfg.categoricals)
        else:
            x = df[modelcfg.features]

        y = df[modelcfg.dependent]

        if modelcfg.dependent_encode:
            y = modelcfg.dependent_encoder.fit_transform(y)

        x_train, x_test, y_train, y_test = train_test_split(
            x, y, test_size=modelcfg.testsize, random_state=0)

        x_train = DatasetPreprocessing.transform_columns(
            modelcfg.imputer.fit_transform, x_train, modelcfg.non_categoricals)
        x_test = DatasetPreprocessing.transform_columns(
            modelcfg.imputer.transform, x_test, modelcfg.non_categoricals)

        if modelcfg.has_standard_scaled():
            if not modelcfg.standard_scaler_fitted:
                x_train = DatasetPreprocessing.transform_columns(
                    modelcfg.standard_scaler.fit_transform, x_train, modelcfg.standard_scaled)
                modelcfg.standard_scaler_fitted = True

            x_test = DatasetPreprocessing.transform_columns(
                modelcfg.standard_scaler.transform, x_test, modelcfg.standard_scaled)

        return modelcfg, x_train, x_test, y_train, y_test

    @staticmethod
    def prepare_prediction(modelcfg, df):
        if modelcfg.has_categoricals():
            x = pd.get_dummies(df[modelcfg.features],
                               columns=modelcfg.categoricals)
        else:
            x = df[modelcfg.features]

        x = x.reindex(columns=modelcfg.columns, fill_value=0)

        x = DatasetPreprocessing.transform_columns(
            modelcfg.imputer.transform, x, modelcfg.non_categoricals)

        if modelcfg.has_standard_scaled():
            x = DatasetPreprocessing.transform_columns(
                modelcfg.standard_scaler.transform, x, modelcfg.standard_scaled)

        return x

    @staticmethod
    def transform_columns(func, df, columns):
        df_temp = df.copy()
        df_temp[columns] = func(df[columns])
        return df_temp



import click

from .service import run

@click.command()
@click.option('--production', is_flag=True, default=False)
@click.option('--host', default='0.0.0.0')
@click.option('--port', default=5000)
def cli(**kwargs):
    run(**kwargs)

if __name__ == '__main__':
    cli(prog_name='learninghouse')


#coding: utf-8

from flask import Flask
from flask_restful import Api

from waitress import serve

from . import __version__

from .model import ModelPrediction, ModelTraining, ModelAPI

print(__version__)

app = Flask(__name__)
api = Api(app)

api.add_resource(ModelPrediction, '/prediction/<string:model>')
api.add_resource(ModelTraining, '/training/<string:model>')
api.add_resource(ModelAPI, '/info/<string:model>')


def run(production, host, port):
    if production:
        serve(app, host=host, port=port)
    else:
        app.run(host=host, port=port, debug=False)



#coding: utf-8

from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

class EstimatorFactory():
    @staticmethod
    def get_estimator(estimatorcfg):
        options = estimatorcfg['options']

        ESTIMATORS = {
            "DecisionTreeClassifier": DecisionTreeClassifier,
            "RandomForestClassifier": RandomForestClassifier
        }

        return ESTIMATORS[estimatorcfg['class']](**options)



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt", "r") as fh:
    requirements = [line.strip() for line in fh]

setuptools.setup(
    name='learninghouse',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='learningHouse - Teach your smart home everything',
    long_description=long_description,
    long_description_content_type='text/markdown',        
    url='https://github.com/LearningHouseService/learninghouse-core',
    author='Johannes Ott',
    author_email='info@johannes-ott.net',
    classifiers=[ 
        'Development Status :: 4 - Beta',

        'Topic :: Home Automation',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='smart home, machine learning, house automation',
    package_dir={'learninghouse': 'learninghouse'},
    packages=['learninghouse'],
    python_requires='>=3.6, <4',
    install_requires = requirements,
    entry_points={
        'console_scripts': [
            'learninghouse=learninghouse.cli:cli'
        ]
    },
    project_urls={
        'Bug Reports': 'https://github.com/LearningHouseService/learninghouse-core/issues'
    }
)


#coding: utf-8

from flask import request, jsonify
from flask_restful import Resource
from werkzeug.exceptions import BadRequest

from joblib import load, dump

import json

from os import path, stat
import sys

import time

import pandas as pd
import numpy as np

from . import __version__

from .preprocessing import DatasetPreprocessing
from .estimator import EstimatorFactory


from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.metrics import confusion_matrix, accuracy_score


class ModelConfiguration():
    CONFIG_FILE = 'models/config/%s.json'
    COMPILED_FILE = 'models/compiled/%s.pkl'

    def __init__(self, name):
        self.name = name

        self.__json_config = self.__load_initial_config(name)

        self.estimatorcfg = self.__required_config('estimator')
        self.testsize = self.__required_config('test_size')
        self.features = self.__required_config('features')
        self.dependent = self.__required_config('dependent')

        self.categoricals = self.__optional_config('categoricals')
        if self.categoricals is None:
            self.non_categoricals = self.features
        else:
            self.non_categoricals = [
                item for item in self.features if item not in set(self.categoricals)]

        self.imputer = SimpleImputer(missing_values=np.nan, strategy='mean')

        self.standard_scaled = self.__optional_config('standard_scaled')
        if self.standard_scaled is not None:
            self.standard_scaler = StandardScaler()
            self.standard_scaler_fitted = False

        self.dependent_encode = self.__optional_config(
            'dependent_encode', False)
        if self.dependent_encode:
            self.dependent_encoder = LabelEncoder()

        self.estimator = None
        self.columns = None
        self.score = 0.0
        self.confusion = None
        self.version = __version__

    def __load_initial_config(self, name):
        with open(ModelConfiguration.CONFIG_FILE % name, 'r') as configFile:
            return json.load(configFile)

    def __required_config(self, param):
        if param in self.__json_config:
            return self.__json_config[param]
        else:
            raise Exception('Missing required param %s' % param)

    def __optional_config(self, param, default=None):
        if param in self.__json_config:
            return self.__json_config[param]
        else:
            return default

    def has_categoricals(self):
        return self.categoricals is not None

    def has_standard_scaled(self):
        return self.standard_scaled is not None

    def has_columns(self):
        return self.columns is not None

    def config_object(self):
        return {
            'name': self.name,
            'estimator_config': self.estimatorcfg,
            'features': self.features,
            'categoricals': self.categoricals,
            'standard_scaled': self.standard_scaled,
            'dependent_encode': self.dependent_encode,
            'dependent': self.dependent,
            'score': self.score,
            'confusion': self.confusion,
            'version': {
                'model': self.version,
                'service': __version__
            }
        }

    def dump(self, estimator, columns, score, confusion):
        self.estimator = estimator
        self.columns = columns
        self.score = score
        self.confusion = confusion

        dump(self, ModelConfiguration.COMPILED_FILE % self.name)


class ModelAPI(Resource):
    @staticmethod
    def make_json_response(data, status_code=200, error_code=None):
        if error_code is not None:
            data['error'] = error_code

        resp = jsonify(data)
        resp.status_code = status_code
        return resp

    @staticmethod
    def get(model):
        try:
            modelcfg = load(ModelConfiguration.COMPILED_FILE % model)
            return ModelAPI.make_json_response(modelcfg.config_object())
        except FileNotFoundError:
            return ModelAPI.make_json_response({}, 404, 'NOT_TRAINED')
        except:
            print("Unexpected error:", sys.exc_info()[0])
            return ModelAPI.make_json_response({}, 500, 'UNKNOWN_ERROR')


class ModelTraining(Resource):
    TRAINING_FILE = 'models/training/%s.csv'

    @staticmethod
    def put(model):
        if request.content_length > 0 and request.is_json:
            filename = ModelTraining.TRAINING_FILE % model
            json_data = request.get_json()
            json_data = DatasetPreprocessing.add_time_information(json_data)
            if path.exists(filename):
                df_temp = pd.read_csv(filename)
                df = df_temp.append([json_data], ignore_index=True)
            else:
                df = pd.DataFrame([json_data])

            df.to_csv(filename, sep=',', index=False)

            return ModelTraining.train(model, df)
        else:
            return ModelAPI.make_json_response({}, 400, 'BAD_REQUEST')

    @staticmethod
    def post(model):
        filename = ModelTraining.TRAINING_FILE % model

        if request.content_length == 0:
            if path.exists(filename):
                df = pd.read_csv(filename)
                return ModelTraining.train(model, df)
            else:
                return ModelAPI.make_json_response({}, 202, 'NOT_ENOUGH_TRAINING_DATA')
        else:
            return ModelAPI.make_json_response({}, 400, 'BAD_REQUEST')

    @staticmethod
    def train(model, df):
        try:
            modelcfg = ModelConfiguration(model)

            if len(df.index) < 10:
                return ModelAPI.make_json_response({}, 202, 'NOT_ENOUGH_TRAINING_DATA')

            modelcfg, x_train, x_test, y_train, y_test = DatasetPreprocessing.prepare_training(
                modelcfg, df)

            estimator = EstimatorFactory.get_estimator(modelcfg.estimatorcfg)

            columns = x_train.columns

            estimator.fit(x_train, y_train)

            y_pred = estimator.predict(x_test)

            cm = confusion_matrix(y_test, y_pred)
            score = accuracy_score(y_test, y_pred)

            modelcfg.dump(estimator, columns, score, cm.tolist())

            return ModelAPI.make_json_response(modelcfg.config_object())
        except FileNotFoundError:
            return ModelAPI.make_json_response({}, 404, 'NO_CONFIGURATION')
        # except:
        #    print("Unexpected error:", sys.exc_info()[0])
        #    return ModelAPI.make_json_response({}, 500, 'UNKNOWN_ERROR')


class ModelPrediction(Resource):
    modelcfgs = {}

    @staticmethod
    def post(model):
        try:
            modelcfg = ModelPrediction._load_modelcfg(model)

            json_data = request.get_json(force=True)
            json_data = DatasetPreprocessing.add_time_information(json_data)
            query = pd.DataFrame([json_data])

            x = DatasetPreprocessing.prepare_prediction(modelcfg, query)

            prediction = modelcfg.estimator.predict(x)

            if modelcfg.dependent_encode:
                prediction = modelcfg.dependent_encoder.inverse_transform(
                    prediction)
                prediction = list(map(bool, prediction))
            else:
                prediction = list(map(float, prediction))

            result = {
                'model': modelcfg.config_object(),
                'prediction': prediction[0]
            }

            return ModelAPI.make_json_response(result)
        except BadRequest as e:
            return ModelAPI.make_json_response({}, 400, 'BAD_REQUEST')
        except KeyError as e:
            return ModelAPI.make_json_response({'message': e.args[0]}, 400, 'MISSING_KEY')
        except FileNotFoundError:
            return ModelAPI.make_json_response({}, 404, 'NOT_TRAINED')
        except:
            print("Unexpected error:", sys.exc_info()[0])
            return ModelAPI.make_json_response({}, 500, 'UNKNOWN_ERROR')

    @staticmethod
    def _load_modelcfg(model):
        filename = ModelConfiguration.COMPILED_FILE % model
        stamp = stat(filename).st_mtime

        if model in ModelPrediction.modelcfgs:
            if ModelPrediction.modelcfgs[model]['stamp'] == stamp:
                modelcfg = ModelPrediction.modelcfgs[model]['model']
            else:
                modelcfg = load(filename)
                ModelPrediction.modelcfgs[model] = {
                    'stamp': stamp,
                    'model': modelcfg
                }
        else:
            modelcfg = load(filename)
            ModelPrediction.modelcfgs[model] = {
                'stamp': stamp,
                'model': modelcfg
            }

        return modelcfg
